#include <gst/gst.h>
#include <gst/interfaces/propertyprobe.h>

#include "audio-sources.h"
#include "pulseaudio.h"
#include "support.h"  // _(x)

// Note: The get_audio_devices() function has two variants.
// This function returns the list of available audio-devices (audio sources) such as sound-cards, microphones and webcams. 
// The 1.st version uses PulseAudio to get the device list. 
// The second variant uses GStreamer, but GStreamer fails to return the device name (it is always NULL, at least in Ubuntu 10.04).
// Later we should avoid the PulseAudio dependency and use GStreamer for all functions. 

void audio_sources_init() {
	;
}

void audio_sources_exit() {
	;
}

DeviceItem *device_item_create(gchar *id, gchar *name) {
	DeviceItem *item = g_malloc0(sizeof(DeviceItem));
	if (id) item->id = g_strdup(id);
    if (name) item->name = g_strdup(name);
    return item;
} 

void device_item_free(DeviceItem *item) {
	if (!item) return;
	if (item->id) g_free(item->id);
	if (item->name) g_free(item->name);
	// Free the DeviceItem node
	g_free(item);
}

gchar *get_device_for_name(gchar *name) {
	// Return device id for the given name/description  

	// Get list of audio devices
	GList *lst = get_audio_devices(DEFAULT_AUDIO_SOURCE);
	GList *n = g_list_first(lst);
	gchar *found_id = NULL;

	while (n) {
		DeviceItem *item = (DeviceItem*)n->data;

		// Names do match?
		if (!g_strcmp0(item->name, name)) {
			found_id = g_strdup(item->id); 	
			break;
		} 

		// Id matches the name/description?
		if (!g_strcmp0(item->id, name)) {
			found_id = g_strdup(item->id); 	
			break;
		} 

		n = g_list_next(n);
    }

	// Free the list
    g_list_foreach(lst, (GFunc)device_item_free, NULL);
    g_list_free(lst); lst = NULL;

	// The caller should g_free() this value 
	return found_id;
} 

GList *add_default_device(GList *dev_list) {
	// Add "Default" device
    // This value can be set in the pavucontrol tool, "Recording" page.
    // See also: http://pulseaudio.org/wiki/DefaultDevice

    DeviceItem *item = device_item_create(NULL, _("System's default device"));
    return g_list_append(dev_list, item);
}


// Choose one of these get_audio_devices() functions

#if 1  
GList *get_audio_devices(gchar *pipe_name) {
	// Use PulseAudio to get the device ids and names.
	// This returns both device id and correct name/description. 
	// Compilation requires -lpulse: pkg-config --cflags --libs libpulse
	GList *dev_list = pa_get_devicelist();
    dev_list = add_default_device(dev_list);
	return dev_list;
}
#endif

#if 0
GList *get_audio_devices(gchar *pipe_name) {

	// Get audio devices (audio sources) from GStreamer.
	// This returns correct device id, but the name/description is NULL. 

    GstElementFactory *factory;
    GstElement *element;

	// Device list
    GList *dev_list = NULL;

    factory = gst_element_factory_find(pipe_name);

    if (!factory) {
        GST_WARNING("Failed to find factory for pipeline %s.\n", pipe_name);
        return dev_list;
    }

    element = gst_element_factory_create(factory, "test");
    const gchar *longname = gst_element_factory_get_longname(factory);
    if (!element) {
        GST_WARNING("Failed to create instance of factory '%s' (%s)", longname, GST_PLUGIN_FEATURE (factory)->name);
        return dev_list;
    }

    const gchar *klass_name = gst_element_factory_get_klass(factory);
    const gchar *description = gst_element_factory_get_description(factory);
    const gchar *author = gst_element_factory_get_author(factory);

    gint n;
    gchar *name;
    GValueArray *array;

    GObjectClass *klass = G_OBJECT_GET_CLASS(element);

    GstPropertyProbe *probe = GST_PROPERTY_PROBE(element);
    const GParamSpec *pspec = gst_property_probe_get_property(probe, "device");

    /* Do we have a "device" property? */
    if (!g_object_class_find_property(klass, "device") || !GST_IS_PROPERTY_PROBE(element) || !probe || !pspec) {
        GST_DEBUG ("Found source '%s' (%s) - no device", longname, GST_PLUGIN_FEATURE (factory)->name);
        return dev_list;
    }

    /* Set autoprobe[-fps] to FALSE to avoid delays when probing. */
    if (g_object_class_find_property(klass, "autoprobe")) {
        g_object_set(G_OBJECT(element), "autoprobe", FALSE, NULL);
    }

    if (g_object_class_find_property(klass, "autoprobe-fps")) {
        g_object_set(G_OBJECT (element), "autoprobe-fps", FALSE, NULL);
    }

    array = gst_property_probe_probe_and_get_values(probe, pspec);
    if (array != NULL) {

        /* default device item, so we can let the element handle it */
        if (array->n_values > 0) {
            g_print("This is the default device (default)\n");
        }

        for (n = 0; n < array->n_values; n++) {
            GValue *device;
            device = g_value_array_get_nth(array, n);
            g_object_set_property (G_OBJECT (element), "device", device);

            if (gst_element_set_state(element, GST_STATE_READY) != GST_STATE_CHANGE_SUCCESS) {
                GST_WARNING("Found source '%s' (%s) - device %s failed to open", longname, GST_PLUGIN_FEATURE (factory)->name, g_value_get_string (device));
                continue;
            }

            g_object_get(G_OBJECT(element), "device-name", &name, NULL);

			// TODO: 
			// The name is always NULL.
			// See Launchpad bug: https://bugs.launchpad.net/ubuntu/+source/gstreamer0.10/+bug/607162

            DeviceItem *item = device_item_create((gchar*)g_value_get_string(device), name);
            dev_list = g_list_append(dev_list, item);

            // g_print("M1: Found source for '%s' (%s) - device %s '%s'\n", longname, GST_PLUGIN_FEATURE (factory)->name, g_value_get_string(device), name);

        }
    }

    gst_element_set_state(element, GST_STATE_NULL);
    gst_object_unref(GST_OBJECT(element));
    gst_object_unref(GST_OBJECT(factory));

    dev_list = add_default_device(dev_list);

    return dev_list;
}
#endif
