#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <math.h>

#include "applet.h"
#include "support.h"
#include "utility.h"
#include "gconf.h"
#include "log.h"
#include "rec-window.h"
#include "gst-recorder.h"
#include "audio-sources.h"
#include "media-profiles.h"
#include "timer.h"

// Reference to the global settings window
extern SettingsWindow g_win;

// Global variable to hold GStreamer variables and data
GstRec g_rec;

#define AUDIO_RECORDER 101
#define AUDIO_RECORDER_ERR1 -1

GstElement *gst_recorder_create_pipeline(gchar *audio_source, gchar *audio_device, gchar *enc_pipeline, 
                                         gchar *output_filename, gboolean append_to_file, GError **error);

static void tick_func_stop();
static void tick_func_exit_cb(gpointer userdata);
static gboolean tick_func_cb(gpointer userdata);

gboolean is_valid_widget(GtkWidget *widget);

// Collect some data from the GStreamer's level element
void level_buf_add(gdouble rms_dB, gdouble peak_dB, gdouble norm_peak, gdouble decay_dB) {
	// Add the values to the circular buffer
	g_rec.level_buf[g_rec.level_buf_pos].rms_dB = rms_dB;       // Not used
	g_rec.level_buf[g_rec.level_buf_pos].peak_dB = peak_dB;     // Not used
	g_rec.level_buf[g_rec.level_buf_pos].norm_peak = norm_peak; // Normalized peak_dB (0.0 - 1.0). Used by timer.c to test "silence".
	g_rec.level_buf[g_rec.level_buf_pos].decay_dB = decay_dB;   // Not used
	g_rec.level_buf_pos = (g_rec.level_buf_pos + 1) % MAX_LEVEL_VALUES;
}

void rec_set_data_lock() {
	// Set the data lock
	g_mutex_lock(g_rec.data_lock);
}

void rec_free_data_lock() {
	// Free the data lock
	g_mutex_unlock(g_rec.data_lock);
}

void gst_recorder_init() {
	// Reset g_rec structure
	g_rec.tick_func_id = 0;
	g_rec.pipeline = NULL;

	// Create mutex, lock. Controls access to g_rec.level_buf[].
	g_rec.data_lock = g_mutex_new();
	g_rec.level_buf_pos = 0;
}

void gst_recorder_exit() {
	g_mutex_free(g_rec.data_lock);
}

gboolean gst_recorder_is_recording() {
	// The pipeline has been created?
	if (!GST_IS_ELEMENT(g_rec.pipeline)) return FALSE;

	// The state is?
	// Ref: http://www.gstreamer.net/data/doc/gstreamer/head/gstreamer/html/GstElement.html#gst-element-get-state
	GstState cur_state, pending;
	gst_element_get_state(g_rec.pipeline, &cur_state, &pending, 0);
	if (cur_state == GST_STATE_NULL) return FALSE;

	return TRUE;
}

gchar *check_folder_name(gchar *folder_name) {

	// Create the folder, this may fail.
	g_mkdir(folder_name, 0755);

	// Directory exists?
	if (!g_file_test(folder_name, G_FILE_TEST_IS_DIR)) {
		g_free(folder_name);
		// Return $HOME
		folder_name = g_strdup(getenv("HOME"));
	}		
	return folder_name;
}

gchar *gst_get_media_profile_id() {
	// Return the saved media profile id. Typical values are: "aac", "mp3", "cdlossless" and "cdlossy".
    // See gconf-editor, key: system -> gstreamer -> 0.10 -> audio -> profiles.
	gchar *id = NULL;
	conf_get_string_value("media_format", &id);

    // Is this id real?
    GMAudioProfile *media_profile = media_profiles_get_profile(id);

    if (!media_profile) {
        g_free(id);
        id = g_strdup("cdlossy"); // = .ogg audio format
    }

    // The caller should g_free() this value
    return id;
}

gchar *gst_generate_filename() {
	// Generate new unique filename

	// Get the saved media profile id (aac, mp3, cdlossless, cdlossy, etc).
    gchar *media_profile_id = gst_get_media_profile_id();

	// .ogg, .mp3, etc
	gchar *file_ext = media_profiles_get_extension(media_profile_id);

	// Last used filename (user may have typed this manually)
	gchar *last_filename = NULL;
	conf_get_string_value("last_filename", &last_filename);
	g_strstrip(last_filename);

	// Save to this folder
	gchar *folder_name = NULL;
	conf_get_string_value("folder_name", &folder_name);
	g_strstrip(folder_name);
	folder_name = check_folder_name(folder_name);

	// Pattern to generate the filename (includes base + date + time pattern)
	gchar *filename_pattern = NULL;
	conf_get_string_value("filename_pattern", &filename_pattern);
	g_strstrip(filename_pattern);
	
	// Pattern cannot have "/" character (edit string inplace)
	g_strdelimit(filename_pattern, "/", '-');

	gboolean append_to_file = FALSE;
	conf_get_boolean_value("append_to_file", &append_to_file);

	LOG_DEBUG("last_filename=%s  pattern=%s  append_to_file=%d\n", last_filename, filename_pattern, append_to_file);

	// Auto generate filename?
	if ((!append_to_file) || str_length(last_filename,1024) < 2)
	{
		// Do we have a proper pattern?
		if (str_length(filename_pattern,128) < 1) {
			g_free(filename_pattern);
			filename_pattern = g_strdup(_("rec-%C%y%m%d-%H:%M:%S"));
			conf_save_string_value("filename_pattern", filename_pattern);
		}

		// Make a filename from the pattern
		g_free(last_filename);
		last_filename = substitute_time_and_date_pattern(filename_pattern);

		// Invalid pattern?
		if (str_length(last_filename, 2048) < 1) {
			// Filename not given, set it to "Some filename"
			last_filename= g_strdup(_("Some filename"));
		}
	}

	// Split the last_filename
	gchar *path=NULL, *basename=NULL, *ext=NULL;
	split_filename3(last_filename, &path, &basename, &ext);

	// Remove old and set real (.ogg, .mp3, etc.) file extension
	gchar *s = g_strdup_printf("%s.%s", basename, file_ext);
	g_free(path);  
	g_free(basename);  
	g_free(ext);

	// Build the entire filename; path + file.ext
	gchar *fname = g_build_filename(folder_name, s, NULL);

	// Can we write to this folder/file?
	if (!is_file_writable(fname)) {
		gchar *msg = g_strdup_printf(_("Cannot write to file %s.\n"), fname);

		LOG_DEBUG("Cannot write to file %s.\n", fname);

		messagebox_error(msg, NULL);
		g_free(msg);
		g_free(fname);
		fname = NULL;
	}

	if (fname) {
		conf_save_string_value("last_filename", s);
	}

	g_free(s);
    g_free(media_profile_id);
	g_free(file_ext);
	g_free(last_filename);
	g_free(folder_name);
	g_free(filename_pattern);

	LOG_DEBUG("Generated filename is:%s.\n", fname);

	return fname;	
}

gboolean gst_recorder_start_recording() {

	LOG_DEBUG("\n--------- gst_recorder_start_recording() ----------\n");

	// Stop any previous tick functions, stop timer function etc
	gst_recorder_stop_recording();

	gchar *filename;
	filename = gst_generate_filename();
	if (!filename)
	{
		gst_recorder_stop_and_reset();
		return FALSE;
	}

	// Get the saved media profile id (aac, mp3, cdlossless, cdlossy, etc).
    gchar *media_profile_id = gst_get_media_profile_id();

    // Get partial pipline from GConf for this media_profile_id
    gchar *encoder_pipeline = media_profiles_get_pipeline(media_profile_id);
    g_free(media_profile_id);

	// Get device name and id
	gchar *device_name = NULL;
	conf_get_string_value("audio_device", &device_name);

	gchar *device_id = get_device_for_name(device_name);
	g_free(device_name);

	gboolean append_to_file;
	conf_get_boolean_value("append_to_file", &append_to_file); // Append to an existing file?

	// Determine audio_source. Typical values are: "pulsesrc", "gconfaudiosrc", "autoaudiosrc" or "alsasrc". 
	gchar *audio_source = DEFAULT_AUDIO_SOURCE;

	gboolean ret = FALSE;

	// User has selected the "Default" device?
	if (device_id == NULL) {
		// So what is this "Default" device/audio-source then?
		// Ref: http://ubuntuforums.org/showpost.php?p=9686252&postcount=15

		// Try first "gconfaudiosrc"
		// Its real value is probably set in the GConf registry. 
		// Start gconf-editor and browse to key /system/gstreamer/0.10/default (audiosrc).
		audio_source = "gconfaudiosrc";
		GstElement *source = gst_element_factory_make(audio_source, "test-audio-source");
		ret = (source != NULL);
		g_object_unref(source);

		if (!ret) {
			// Try "autoaudiosrc"
			audio_source = "autoaudiosrc";
			source = gst_element_factory_make(audio_source, "test-audio-source");
			ret = (source != NULL);
			g_object_unref(source);
		}

		if (!ret) {
			audio_source = DEFAULT_AUDIO_SOURCE;
		}
	}

	// Now build the GStreamer pipeline with these values
	GError *error = NULL;
	g_rec.pipeline = gst_recorder_create_pipeline(audio_source, device_id, encoder_pipeline, filename, append_to_file, &error);

    ret = TRUE;

	// Ok?
	if (error) {
		// Errors.

		// Show error message
		messagebox_error(error->message, NULL);		

		// Stop and reset everything
		gst_recorder_stop_and_reset();

        g_error_free(error);

        ret = FALSE;
	}
    else {
		// Call the callback function with a NULL value. It resets its internal static variables and exits.
		tick_func_cb(NULL);

		// Start the tick (timer) function, the call intervall is 1 second.
		g_rec.tick_func_id = g_timeout_add_seconds_full(G_PRIORITY_DEFAULT, 1, (GSourceFunc)tick_func_cb, g_strdup(filename), (GDestroyNotify)tick_func_exit_cb);
		// Note: The tick_func_exit_cb() function will g_free() the filename.

		// Start a timer function to process "Stop after", "Stop at" settings. 
		timer_func_start();
    }		

    g_free(encoder_pipeline);
	g_free(device_id);
	g_free(filename);
	return ret;
}

gboolean gst_recorder_stop_and_reset() {
	// Stop recording 
	gst_recorder_stop_recording();

	// Reset window and its widgets
	win_set_widgets();

	// Reset aplet and its widgets
	applet_set_widgets();

	return TRUE;
}

static void gst_consume_events()
{
    GstBus *bus = gst_element_get_bus(GST_ELEMENT(g_rec.pipeline));

    GstMessage *message = NULL;
    while (1) {
        message = gst_bus_poll(bus, GST_MESSAGE_ANY, GST_SECOND*4/*timeout*/);
        if (!message) return;

        switch (message->type) {

        case GST_MESSAGE_EOS:
            LOG_DEBUG("Detected GST_MESSAGE_EOS. Very good, the data flow has terminated properly.\n");
            gst_message_unref(message);
            return;

        case GST_MESSAGE_WARNING:
            LOG_DEBUG("Detected GST_MESSAGE_WARNING.\n");

        case GST_MESSAGE_ERROR:{
            if (message->type == GST_MESSAGE_ERROR)
                LOG_DEBUG("Detected GST_MESSAGE_ERROR.\n");

            GError *error = NULL;
            gchar *debug = NULL;
            gst_message_parse_error(message, &error, &debug);
            gst_object_default_error(GST_MESSAGE_SRC(message), error, debug);
            gst_message_unref(message);

            if (error)  {
                LOG_ERROR("Error message:", error->message);
                g_error_free(error);
            }    

            g_free(debug);
            return;
        }
        default:
            gst_message_unref(message);
            break;
       }
    }
}

gboolean gst_recorder_stop_recording() {

	// Stop the tick_func_cb() function
	tick_func_stop();

	// Stop the timer function
	timer_func_stop();

	// Terminate pipeline
	if (!GST_IS_PIPELINE(g_rec.pipeline)) return TRUE;

    LOG_DEBUG("gst_recorder_stop_recording()\n");

    // Shutdown the _source_ element. This will send an EOS event.
    // This is the same as calling gst-launch with -e option.
    // Sending EOS will terminate the stream/file properly. This is very important for ACC (.m4a) files.

    // Get the "audio-source" element (by name)
    GstElement *source = gst_bin_get_by_name(GST_BIN(g_rec.pipeline), "audio-source");
    if (GST_IS_ELEMENT(source)) {
        gst_element_set_state(source, GST_STATE_NULL);
    }
    gst_object_unref(GST_OBJECT(source));

    g_usleep(GST_USECOND * 1);

    // Consume events, mainly wait for GST_MESSAGE_EOS event to occur.
    gst_consume_events();

    // Shutdown the entire pipeline
    gst_element_set_state(g_rec.pipeline, GST_STATE_NULL);

    // Destroy it
    gst_object_unref(GST_OBJECT(g_rec.pipeline));
	g_rec.pipeline = NULL;

	LOG_DEBUG("--------- Pipeline closed and destroyed ----------\n\n");

	return TRUE;
}


gchar *gst_get_output_filename() {
    // Return the current output filename

    // Valid pipeline?
    if (!GST_IS_ELEMENT(g_rec.pipeline)) return NULL;

    // Get the "output-sink" element (by name)
    GstElement *filesink = gst_bin_get_by_name(GST_BIN(g_rec.pipeline), "output-sink");

    // Got filesink?
    if (!GST_IS_ELEMENT(filesink)) return NULL;

    gchar *filename = NULL;
    g_object_get(G_OBJECT(filesink), "location", &filename, NULL);
    g_object_unref(filesink);

    // The caller should g_free() the value
    return filename;
}

static gboolean level_message_handler_cb(GstBus *bus, GstMessage *message, void *user_data)
{
	if (message->type == GST_MESSAGE_ELEMENT) {
		const GstStructure *s = gst_message_get_structure (message);
		const gchar *name = gst_structure_get_name (s);

		if (g_str_equal (name, "level")) {
			gint channels;
			const GValue *list;
			const GValue *value;

			gint i;

			// We can get the number of channels as the length of any of the value lists
			// See: http://www.gstreamer.net/data/doc/gstreamer/head/gst-plugins-good-plugins/html/gst-plugins-good-plugins-level.html
			list = gst_structure_get_value(s, "rms");
			channels = gst_value_list_get_size(list);
			gdouble rms_dB, decay_dB, peak_dB, norm_peak;

			for (i = 0; i < channels; ++i) {
				list = gst_structure_get_value (s, "rms");
				value = gst_value_list_get_value (list, i);
				rms_dB = g_value_get_double(value);

				list = gst_structure_get_value (s, "peak");
				value = gst_value_list_get_value (list, i);
				peak_dB = g_value_get_double(value);

				norm_peak = exp(peak_dB / 20);
				if (norm_peak > 1.0) norm_peak = 1.0;

				list = gst_structure_get_value(s, "decay");
				value = gst_value_list_get_value (list, i);
				decay_dB = g_value_get_double (value);

				// Save the level values so other threads can analyze the values (eg. test for "silence" or "talk/music")
				// Set lock
				rec_set_data_lock();

				// Add the value to the buffer
				level_buf_add(rms_dB, peak_dB, norm_peak, decay_dB);

				// Free the lock
				rec_free_data_lock();

				// Set progressbar value
				if (GTK_IS_PROGRESS_BAR(g_win.level_bar))  {
					gtk_progress_bar_set_fraction(GTK_PROGRESS_BAR(g_win.level_bar), norm_peak * 0.96); // (0.96, just cheating a bit)
				}
			}

		}
	}

	// TRUE: Continue calling this function
	return TRUE;
}

static void record_state_changed_cb(GstBus *bus, GstMessage *msg, void *userdata)
{
	GstState  new_state;
	gst_message_parse_state_changed(msg, NULL, &new_state, NULL);

	// We are only interested in state changes of the top-level pipeline.
	if (msg->src != GST_OBJECT(g_rec.pipeline)) return;

	LOG_DEBUG("Pipeline state changed to: %s.\n", gst_element_state_get_name(new_state));

	switch (new_state) {
	case GST_STATE_PLAYING:
		;
		break;
	case GST_STATE_READY:
		;
	case GST_STATE_PAUSED:
		;
		break;
	default:
		break;
	}
}

static void pipeline_error_cb(GstBus *bus, GstMessage *msg, void *userdata)
{
	GError *error = NULL;
	gchar *dbg = NULL;

	gst_message_parse_error(msg, &error, &dbg);
	g_return_if_fail(error != NULL);

	LOG_DEBUG("\nGot pipeline error: %s.\n", error->message);

	// pipeline = gst_element_get_toplevel(GST_ELEMENT (msg->src));
	if (error->code == GST_RESOURCE_ERROR_BUSY) {
			g_error_free(error);
			g_free(dbg);
			return;
	}

	if (error)
		g_error_free(error);

	g_free(dbg);

	// Maker sure recording stops. Destroy the pipeline if alive
	gst_recorder_stop_and_reset();
}

static void record_eos_msg_cb(GstBus *bus, GstMessage *msg, void *userdata)
{
	LOG_DEBUG("Got EOS. Finishing recording.\n");
}

GstElement *gst_recorder_create_pipeline(gchar *audio_source, gchar *audio_device, gchar *enc_pipeline, gchar *output_filename, gboolean append_to_file, GError **error) {
	// Create a GStreamer pipeline from the given arguments. 
    // Return the pipeline.
	// 
	// audio_source: most likely "pulsesrc" or "alsasrc".
	// 

    GstElement *pipeline = NULL;

    *error = NULL;

	LOG_DEBUG("Create pipeline. The parameters are:\n");
	LOG_DEBUG("\taudio_source=%s\n", audio_source);
	LOG_DEBUG("\taudio_device=%s\n", audio_device);
	LOG_DEBUG("\tencoder_pipeline (from GConf)=%s\n",enc_pipeline);
	LOG_DEBUG("\tfilename=%s\n", output_filename);
	LOG_DEBUG("\tappend_to_file=%s\n", (append_to_file ? "TRUE" : "FALSE"));

	// Some simplified samples of GStreamer pipelines
	// OGG audio output:
	// $ gst-launch-0.10 pulsesrc ! queue ! audioconvert ! vorbisenc ! oggmux ! filesink location=test.oga
	// 
	// FLAC:
	// $ gst-launch-0.10 pulsesrc ! queue ! audioconvert ! flacenc ! filesink location=test.flac
	//
	// WAV:
	// $ gst-launch-0.10 pulsesrc ! queue ! audioconvert ! wavenc ! filesink location=test.wav
	// 
	// MP3:
	// $ gst-launch-0.10 pulsesrc ! queue ! audioconvert ! lame ! filesink location=test.mp3

	// Carefully check for errors 
	gchar *err_msg = NULL;

    gchar *pipeline_cmd = g_strdup_printf ("%s name=audio-source"
                                   " ! level name=level"
                                   " ! queue"
                                   " ! audioconvert"
                                   " ! %s"
                                   " ! filesink name=output-sink",
                                   audio_source,
                                   enc_pipeline);


    LOG_DEBUG("Going to run pipeline: \"%s\"\n", pipeline_cmd);

    GError *err = NULL;
    pipeline = gst_parse_launch(pipeline_cmd, &err);
    g_free(pipeline_cmd);

    if (err) {
        err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s\n"), err->message);
        g_error_free(err);
        goto LBL_1;
    }

    // Get "audio-source" element
    GstElement *source = gst_bin_get_by_name(GST_BIN(pipeline), "audio-source"); 

    if (!GST_IS_ELEMENT(source)) {
        err_msg = g_strdup_printf(_("Cannot find audio element %s.\n"), audio_source);
        goto LBL_1;
    }

	// And set audio device
	if (audio_device)
		g_object_set(G_OBJECT(source), "device", audio_device, NULL);

    g_object_unref(source);

    // Get "output-sink" element 
    GstElement *filesink = gst_bin_get_by_name(GST_BIN(pipeline), "output-sink"); 

    if (!GST_IS_ELEMENT(filesink)) {
        err_msg = g_strdup_printf(_("Cannot find audio element %s.\n"), "filesink");
        goto LBL_1;
    }

    // And set location (filename) and append mode
    g_object_set(G_OBJECT(filesink), "location", output_filename, NULL);
    g_object_set(G_OBJECT(filesink), "append", append_to_file, NULL);

    g_object_unref(filesink);

	// Add a message handler
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));

    // gst_bus_add_watch (bus, bus_info_cb, NULL);
	gst_bus_add_signal_watch(bus);

	// Detect state changes
	g_signal_connect(bus, "message::state-changed", G_CALLBACK(record_state_changed_cb), NULL);

	// Monitor sound level/amplitude
  	g_signal_connect(bus, "message::element", G_CALLBACK(level_message_handler_cb), NULL);

	// Catch error messages
	g_signal_connect(bus, "message::error", G_CALLBACK(pipeline_error_cb), NULL);

	// EOS
	g_signal_connect(bus, "message::eos", G_CALLBACK(record_eos_msg_cb), NULL);

    gst_object_unref(bus);


    // Assume all OK
	gboolean ret = TRUE;

	// Set the pipeline to "playing" state
    if (gst_element_set_state(pipeline, GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE) {
		err_msg = g_strdup(_("Cannot start reading from the stream/pipeline.\n"));
		ret = FALSE;
	} else {
		LOG_DEBUG("Pipeline is OK. Starting recording to %s.\n", output_filename);
	}

	// Are we finally ok?
	if (ret) {
        return pipeline;
    }
	
 LBL_1:
		// Set error object
		if (!err_msg)
			err_msg = g_strdup_printf(_("Cannot create audio pipeline. %s\n"), "");

		LOG_ERROR(err_msg);

		gst_object_unref(GST_OBJECT(pipeline));

		g_set_error(error,
					AUDIO_RECORDER,                 /* Error domain */
                   	AUDIO_RECORDER_ERR1,            /* Error code */
                   	err_msg, "");

		g_free(err_msg);

        // Error
		return NULL;
}

gint64 gst_get_stream_time() {
	// Get time value
	GstFormat format = GST_FORMAT_TIME;
	gint64 val = -1;
	gint secs = 0L;

    // Valid pipeline?
    if (!GST_IS_ELEMENT(g_rec.pipeline)) return 0L;

	if (gst_element_query_position(g_rec.pipeline, &format, &val) && val != -1) {
		secs = val / GST_SECOND;
	}
	return secs;
}

static void tick_func_stop() {
	LOG_DEBUG("tick_func_stop()\n");

	if (g_rec.tick_func_id > 0) {
		g_source_remove(g_rec.tick_func_id); 
		g_rec.tick_func_id = 0;
	}
}

static void tick_func_exit_cb(gpointer userdata){
	// Called when tick_func_cb() terminates.

	LOG_DEBUG("tick_func_exit_cb(), free filename\n");
	
	// g_free() the filename used by tick_func_cb()
	gchar *filename = (gchar*)userdata;

	g_free(filename);
}

static gboolean tick_func_cb(gpointer userdata)
{
	// This timed routine will update the recording time (g_win.time_label) value and file size (g_win.time_label).

	#define UPDATE_SIZE_FREQ 5

	static gulong counter = 0;

	gchar *filename = (gchar*)userdata; // This thread's exit function will g_free() this.

	// Called with NULL filename?
	if (!filename) {
		// Simply reset the counter and return
		counter = 0;
		return TRUE;
	}

	GstFormat format = GST_FORMAT_TIME;
	gint64 val = -1;
	gint secs;

	// Are we still recording?
	gboolean ret = gst_recorder_is_recording();
	if (!ret) return FALSE; 
	
	gchar *time_txt;
	time_txt = NULL;

	// The g_time_label widget is valid?
	ret = is_valid_widget(g_win.time_label);
	if (ret) {
		// Get time value
		if (gst_element_query_position(g_rec.pipeline, &format, &val) && val != -1) {

			secs = val / GST_SECOND;

			guint hours = (guint)(secs / 3600);
			secs = secs - (hours*3600);

			// Count only to 23 hours
			// if (hours > 99) hours = 23;
		
			guint minutes = (guint)(secs / 60);
			guint seconds = secs - (minutes*60);

			time_txt = g_strdup_printf("%02d:%02d:%02d", hours, minutes, seconds); 
		}
	}

	gchar *size_txt;
	size_txt = NULL;

	// Update file size each UPDATE_SIZE_FREQ seconds (or at the very beginning)
	if (counter % UPDATE_SIZE_FREQ == 0 || counter < 5) {

		// g_size_label widget is valid?
		ret = is_valid_widget(g_win.size_label);
		if (ret) {
			// Get file size
			struct stat fstat;
			if (stat(filename, &fstat))
				LOG_ERROR("Cannot get file information of %s.\n", filename);
			else
				size_txt = format_file_size(fstat.st_size);
		} 
	}		

	if (time_txt || size_txt) {
		// Called from a thread, be safe!
		gdk_threads_enter();

		if (time_txt)
			gtk_label_set_text(GTK_LABEL(g_win.time_label), time_txt);

		if (size_txt)
			gtk_label_set_text(GTK_LABEL(g_win.size_label), size_txt);

		gdk_threads_leave();
	}

	g_free(time_txt);
	g_free(size_txt);

	counter += 1;

	// Continue calling this routine
	return TRUE;
}

gboolean is_valid_widget(GtkWidget *widget) {
	// Test if the widget is valid.
	// Possibly called from a thread, be safe!
	gdk_threads_enter();
	gboolean ret = GTK_IS_WIDGET(widget);
	gdk_threads_leave();
	return ret;
}


