/* 
 * Copyright (c) The entire Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include "log.h"
#include "about.h"
#include "gconf.h"
#include "applet.h" 
#include "rec-window.h"
#include "gst-recorder.h"
#include "audio-sources.h"
#include "media-profiles.h"
#include "timer.h"
#include "support.h"

// Note:
// This application has three important global structures/variables.
//  AppletRec g_applet  : Defined in applet.h and main.c, contains widgets of the panel-applet itself. 
//  SettingsWindow g_win: Defined in rec-window.[ch], contains widgets of the settings window.
//  GstRec g_rec        : Defined in gst-recorder.[ch], holds the GStreamer pipeline and data. 

// The window and system settings are saved in the GConf registry.
// Start gconf-editor and browse to apps -> audio_recorder_applet. 

// Note:
// Uncomment this to test this code in an ordinary GTK-window (instead of a panel applet)
// #define AUDIO_REC_APPLET_DEBUG

// Global AppletRec structure
AppletRec g_applet;

// Create applet, its widgets
static void applet_create_content(PanelApplet *applet);

// Menu callbacks
static void applet_about_dialog_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb);
static void applet_start_recording_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb);
static void applet_stop_recording_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb);

// Popup menu on right-click  
static const char right_click_menu_xml [] =
   "<popup name=\"button3\">\n"
   "  <menuitem name=\"Start Recording\" "
   "            verb=\"StartRecording\" "
   "            _label=\"_Start Recording\"/>\n"
   "  <menuitem name=\"Stop Recording\" "
   "            verb=\"StopRecording\" "
   "            _label=\"Sto_p Recording\"/>\n"
   "  <separator/>\n"
   "  <menuitem name=\"About Item\" "
   "            verb=\"MenuAbout\" "
   "            _label=\"_About\"\n"
   "            pixtype=\"stock\" "
   "            pixname=\"gtk-about\"/>\n"
   "</popup>\n";

static const BonoboUIVerb applet_menu_verbs [] = {
        BONOBO_UI_VERB ("MenuAbout", applet_about_dialog_cb),
        BONOBO_UI_VERB ("StartRecording", applet_start_recording_cb),
        BONOBO_UI_VERB ("StopRecording", applet_stop_recording_cb),
        BONOBO_UI_VERB_END
}; 

void applet_set_widgets() {
	// Are we recording?
	gboolean recording = gst_recorder_is_recording();

	gchar *image_file = NULL;

	if (recording) {
		// Red circle (means ON)
		image_file = (gchar*)get_image_path("recording-on.png");
	} else {
		// Green circle (means OFF)
		image_file = (gchar*)get_image_path("recording-off.png");
	}

	// Set image
	if (GTK_IS_IMAGE(g_applet.image)) {
		gtk_image_set_from_file(GTK_IMAGE(g_applet.image), image_file);  	
	}

	g_free(image_file);

    // Set right-click-menu
    BonoboUIComponent *popup_component;
    popup_component = panel_applet_get_popup_component(g_applet.applet);

    if (popup_component) {
        gchar *on = "1";
        gchar *off = "0";
        bonobo_ui_component_set_prop(popup_component, "/commands/StartRecording", "hidden", (recording ? on : off),  NULL);
        bonobo_ui_component_set_prop(popup_component, "/commands/StopRecording", "hidden", (recording ? off : on),  NULL);
    }

}

static void applet_start_recording_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb)
{
	LOG_DEBUG("applet_start_recording_cb()\n");
	win_start_recording_cb(NULL, NULL);
}

static void applet_stop_recording_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb)
{
	LOG_DEBUG("applet_stop_recording_cb()\n");
	win_start_recording_cb(NULL, NULL);
}

static void applet_about_dialog_cb(BonoboUIComponent *component, gpointer user_data, const gchar *verb)
{
	LOG_DEBUG("applet_about_dialog_cb()\n");
	about_this_app();
}

void panel_orient_changed_cb(PanelApplet *applet, guint arg1, gpointer user_data) {
	LOG_DEBUG("panel_orient_changed_cb()\n");

	// The panel's orientation changed, re-create the applet/widgets
	applet_create_content(applet);
}

void panel_size_changed_cb(PanelApplet *panelapplet, gint arg1, gpointer user_data) {
	LOG_DEBUG("panel_size_changed_cb()\n");

	// The panel's size changed.
}

void change_background_cb(PanelApplet *applet, PanelAppletBackgroundType type, GdkColor *color, GdkPixmap *pixmap, gpointer *data) {
	GtkRcStyle *rc_style;
	GtkStyle *style;

	LOG_DEBUG("change_background_cb()\n");

	// Panel's background changed.

	// Reset style
	gtk_widget_set_style(GTK_WIDGET(applet), NULL);
	rc_style = gtk_rc_style_new();
	gtk_widget_modify_style(GTK_WIDGET(applet), rc_style);
	gtk_rc_style_unref(rc_style);

	switch(type) {
	case PANEL_COLOR_BACKGROUND:
		gtk_widget_modify_bg(GTK_WIDGET(applet), GTK_STATE_NORMAL, color);
		break;

	case PANEL_PIXMAP_BACKGROUND:
		style = gtk_style_copy(GTK_WIDGET(applet)->style);
		if (style->bg_pixmap[GTK_STATE_NORMAL]) {
			g_object_unref(style->bg_pixmap[GTK_STATE_NORMAL]);
		}
		style->bg_pixmap[GTK_STATE_NORMAL] = g_object_ref(pixmap);
		gtk_widget_set_style(GTK_WIDGET(applet), style);
		g_object_unref(style);
		break;

	case PANEL_NO_BACKGROUND:
		break;

	default:
		break;
	}
}

void style_set_cb(GtkWidget *widget,GtkStyle *old_style, gpointer data) {
	LOG_DEBUG("style_set_cb()\n");
}

static gboolean applet_label_click_cb(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	LOG_DEBUG("applet_label_click_cb()\n");

	if (event->button != 1) return FALSE;

	// Show settings window
	window_create();
	return TRUE;
}

static void applet_destroy_cb(GtkWidget *widget, gpointer data) {
	LOG_DEBUG("applet_destroy_cb()\n");

	// Stop recording if active
	gst_recorder_stop_and_reset();

	// Clean up gst_recorder.c module
	gst_recorder_exit();

	// Clean up audio_sources.c module
	audio_sources_exit();

	// Clean up timer.c module
	timer_module_exit();

    // Clean up media-profiles.c module
    media_profiles_exit();

	gtk_widget_destroy(widget);
}

static void applet_create_content(PanelApplet *applet) {
	LOG_DEBUG("applet_create_content()\n");

	panel_applet_set_background_widget(PANEL_APPLET(applet), GTK_WIDGET(applet));

	// Save applet ref in a global variable	
	g_applet.applet = applet;

	// Hide the settings window (if open)
	window_hide();

	// Create REC button and a small image
	// [REC>][IMG]

	// Panel orientation is?
	gint panel_orientation = panel_applet_get_orient(applet);

	// Destroy the previous widget setting
	if (GTK_IS_WIDGET(g_applet.box)) {
		gtk_widget_destroy(g_applet.box);
	}

	// Vertical or horizontal setting?
	if ((PanelAppletOrient)panel_orientation ==  PANEL_APPLET_ORIENT_UP || 
        (PanelAppletOrient)panel_orientation ==  PANEL_APPLET_ORIENT_DOWN) {
		// Horizontal top/bottom panel
		g_applet.box = gtk_hbox_new(FALSE, 1); 
	}
	else {
		// Vertical left/right edge panel 
		g_applet.box = gtk_vbox_new(FALSE, 1); 
	}

	// Add a "REC>" label.
	// Translators: Probably no need to translate this label. English is just fine.
	GtkWidget *label = gtk_label_new(_("REC>")); 
	gtk_box_pack_start(GTK_BOX(g_applet.box), label, FALSE, FALSE, 0); 

	// Add image.
	g_applet.image = gtk_image_new();
	gtk_box_pack_start(GTK_BOX(g_applet.box), g_applet.image, FALSE, FALSE, 0); 

	// Set the image, red or green dot. We may be recording ATM.
	applet_set_widgets();

	gtk_container_add(GTK_CONTAINER(applet), g_applet.box);
	g_signal_connect(G_OBJECT(applet), "button_press_event", G_CALLBACK(applet_label_click_cb), NULL);

	gtk_widget_show_all(GTK_WIDGET(applet));
}

static gboolean audio_recorder_applet_fill(PanelApplet *applet, const gchar *iid, gpointer data)
{
	if (strcmp (iid, "OAFIID:AudioRecorderApplet") != 0) 
        return FALSE;

    // gtk_window_set_title(GTK_WINDOW(applet), "Audio Recorder Applet");

#ifdef ENABLE_NLS
    // This is needed for the applet's right-click menu.
	bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
	textdomain(GETTEXT_PACKAGE);
#endif
	gtk_set_locale();

	// Init gst_recorder.c module
	gst_init(NULL, NULL);
	gst_recorder_init();

	// Init audio_sources.c module
	audio_sources_init();

	// Init timer.c module
	timer_module_init();

    // Init and load supported audio formats (MP3, OGG, Flac, etc.) and their pipelines
    media_profiles_init();

	// Show button images (normally not shown in the GNOME)
	// See also gconf-editor, setting '/desktop/gnome/interface'
	g_object_set(gtk_settings_get_default(), "gtk-button-images", TRUE, NULL);

	// Create the applet/widgets
	applet_create_content(applet);

    // Applet properties
	panel_applet_set_flags(PANEL_APPLET(applet), PANEL_APPLET_EXPAND_MINOR);
	// "Show handle" flag: PANEL_APPLET_HAS_HANDLE;

    g_signal_connect(G_OBJECT(applet), "change-orient", G_CALLBACK(panel_orient_changed_cb), NULL);
    g_signal_connect(G_OBJECT(applet), "change-size", G_CALLBACK(panel_size_changed_cb), NULL);
    g_signal_connect(G_OBJECT(applet), "destroy", G_CALLBACK(applet_destroy_cb), NULL);
    g_signal_connect(G_OBJECT(applet), "style-set", G_CALLBACK(style_set_cb), NULL);
    g_signal_connect(G_OBJECT(applet), "change_background", G_CALLBACK(change_background_cb), NULL);

    // These strings are shown when user adds new applets to the GNOME's toolbar panel.
    // Note: Translate and copy these 2 text strings into "data/rec-applet.server.in.in" file.
    gchar *label = _("Audio Recorder Applet");
    label = _("Easy-to-use audio recording tool");

    // Menu items on applet's right-click menu
    label = _("_Start Recording");
    label = _("Sto_p Recording");
    label = _("_About");

    // Setup menu
    panel_applet_setup_menu(PANEL_APPLET(applet), right_click_menu_xml, applet_menu_verbs, NULL);

    BonoboUIComponent *popup_component;
    popup_component = panel_applet_get_popup_component(applet);
    bonobo_ui_component_set_translate(popup_component, "/", right_click_menu_xml, NULL);

    // Hide "Stop Recording" item initially
    if (popup_component) {
        bonobo_ui_component_set_prop(popup_component, "/commands/StopRecording", "hidden", "1",  NULL);
    }

    // Accessibility support
    AtkObject *atk_widget = (AtkObject*)gtk_widget_get_accessible(GTK_WIDGET(applet));
    if (GTK_IS_ACCESSIBLE(atk_widget)) {
        atk_object_set_name (atk_widget, _("Audio Recorder Application"));
        atk_object_set_description(atk_widget, _("Easy-to-use audio recording tool"));
    }

	return TRUE;
}

// Run the applet on the panel? 
#ifndef AUDIO_REC_APPLET_DEBUG

PANEL_APPLET_BONOBO_FACTORY ("OAFIID:AudioRecorderApplet_Factory",
                             PANEL_TYPE_APPLET,
			     			 PACKAGE, 
			     			 PACKAGE_VERSION, 
                             audio_recorder_applet_fill,
                             NULL);


#else
// Run this app in an ordinary GTK-window (much easier to test, debug and develope). 
// Bonobo may spit some warnings, that's ok.
// Run it like: 
// $ src/rec-applet

int main (int argc, char *argv[])
{
	GtkWidget *window;
	GtkWidget *applet;

#ifdef ENABLE_NLS
	bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
	textdomain(GETTEXT_PACKAGE);
#endif
	gtk_set_locale();

	// Initialize
	gtk_init(&argc, &argv);

	bonobo_init(&argc, argv);

	// An ordinary GTK-window
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size(GTK_WINDOW(window), 150, 60);  
	g_signal_connect (window, "destroy", G_CALLBACK (gtk_main_quit), NULL);

	applet = panel_applet_new(); 

	gtk_container_set_border_width(GTK_CONTAINER(window), 12);

	gtk_widget_reparent(applet, window);

	audio_recorder_applet_fill(PANEL_APPLET(applet), "OAFIID:AudioRecorderApplet", NULL);

	gtk_widget_show_all(window);

	gtk_main ();

	return 0;
}

#endif


