#include "pulseaudio.h"
#include "audio-sources.h"

#include "support.h" // _(x)

// Original version of this code is from:
// http://pulseaudio.org/wiki/SampleAsyncDeviceList

// This structure protects the GList when sent through callback functions.
// The list pointer may change, but the holder pointer remains the same.
typedef struct {
	GList *list;
} ListHolder;

static void pa_state_cb(pa_context *c, void *userdata);
static void pa_sinklist_cb(pa_context *c, const pa_sink_info *l, int eol, void *userdata);
static void pa_sourcelist_cb(pa_context *c, const pa_source_info *l, int eol, void *userdata);

GList *pa_get_devicelist() {
    // Define our pulse audio loop and connection variables
    pa_mainloop *pa_ml = NULL;
    pa_mainloop_api *pa_mlapi = NULL;
    pa_operation *pa_op = NULL;
    pa_context *pa_ctx = NULL;

    // We'll need these state variables to keep track of our requests
    int state = 0;
    int pa_ready = 0;

	ListHolder list_holder;
	list_holder.list = NULL;

    // Create a mainloop API and connection to the default server
    pa_ml = pa_mainloop_new();
    pa_mlapi = pa_mainloop_get_api(pa_ml);
    pa_ctx = pa_context_new(pa_mlapi, "test");

    // This function connects to the pulse server
    pa_context_connect(pa_ctx, NULL, 0, NULL);

    // This function defines a callback so the server will tell us it's state.
    // Our callback will wait for the state to be ready.  The callback will
    // modify the variable to 1 so we know when we have a connection and it's
    // ready.
    // If there's an error, the callback will set pa_ready to 2
    pa_context_set_state_callback(pa_ctx, pa_state_cb, &pa_ready);

	gint result = 0;

    // Now we'll enter into an infinite loop until we get the data we receive
    // or if there's an error
    while (!result) {
        // We can't do anything until PA is ready, so just iterate the mainloop
        // and continue
        if (pa_ready == 0) {
            pa_mainloop_iterate(pa_ml, 1, NULL);
            continue;
        }

        // We couldn't get a connection to the server, so exit out
        if (pa_ready == 2) {
            pa_context_disconnect(pa_ctx);
            pa_context_unref(pa_ctx);
            pa_mainloop_free(pa_ml);
			result = -1;
			break;
        }

        // At this point, we're connected to the server and ready to make requests
        switch (state) {
            // State 0: we haven't done anything yet
            case 0:
				// NULL == we do not collect sink (output) devices.
                pa_op = pa_context_get_sink_info_list(pa_ctx, pa_sinklist_cb, NULL);

                // Update state for next iteration through the loop
                state++;
                break;

            case 1:
                // Now we wait for our operation to complete. 
                if (pa_operation_get_state(pa_op) == PA_OPERATION_DONE) {
                    pa_operation_unref(pa_op);

                    // Get the source (input devices) list. We pass a pointer to our list_holder.
                    pa_op = pa_context_get_source_info_list(pa_ctx, pa_sourcelist_cb, &list_holder);

                    // Update the state so we know what to do next
                    state++;
                }
                break;

            case 2:
                if (pa_operation_get_state(pa_op) == PA_OPERATION_DONE) {
                    // Now we're done, clean up and disconnect
                    pa_operation_unref(pa_op);
                    pa_context_disconnect(pa_ctx);
                    pa_context_unref(pa_ctx);
                    pa_mainloop_free(pa_ml);
                   	result = 1;
                }
                break;

            default:
                // We should never see this state
                fprintf(stderr, "Error state %d\n", state);
                result = -1;
        }

		if (result != 0) break;

        // Iterate the main loop and go again.  The second argument is whether
        // or not the iteration should block until something is ready to be
        // done.  Set it to zero for non-blocking.
        pa_mainloop_iterate(pa_ml, 1, NULL);
    }

	// Return the device list
	return list_holder.list;
}

// This callback gets called when our context changes state.  We really only
// care about when it's ready or if it has failed
static void pa_state_cb(pa_context *c, void *userdata) {
        pa_context_state_t state;
        int *pa_ready = userdata;

        state = pa_context_get_state(c);
        switch  (state) {
                // There are just here for reference
                case PA_CONTEXT_UNCONNECTED:
                case PA_CONTEXT_CONNECTING:
                case PA_CONTEXT_AUTHORIZING:
                case PA_CONTEXT_SETTING_NAME:
                default:
                        break;
                case PA_CONTEXT_FAILED:
                case PA_CONTEXT_TERMINATED:
                        *pa_ready = 2;
                        break;
                case PA_CONTEXT_READY:
                        *pa_ready = 1;
                        break;
        }
}

// pa_mainloop will call this function when it's ready to tell us about a sink.
static void pa_sinklist_cb(pa_context *c, const pa_sink_info *l, int eol, void *userdata) {
	// We do not collect these devices.
}

// See above. This callback is pretty much identical to the previous.
static void pa_sourcelist_cb(pa_context *c, const pa_source_info *l, int eol, void *userdata) {
	// Collect input devices (audio sources).

	ListHolder *list_holder = (ListHolder*)userdata;

    if (eol > 0) {
        return;
    }

	// Replace the "Monitor of" prefix with "(speaker)" word
    // TODO: 
    // Test the device's "device.class='monitor'" value instead of the description.
	gchar *s = NULL;
	if (g_strrstr(l->description, "Monitor of") == l->description) {
		s = g_strdup(l->description + 10);

		// Remove spaces
		g_strstrip(s);

		// Add word "(Audio output)" to the tail
		// Translators: "(Audio output)" is normally a loudspeaker that produces sound.
		gchar *s2 = g_strdup_printf("%s %s", s, _("(Audio output)"));
		g_free(s);
		s = s2;

	} else {
		s = g_strdup(l->description);
	}

	// Create new DeviceItem and insert it to the list
	DeviceItem *item = device_item_create((gchar*)l->name, s);
	list_holder->list = g_list_append(list_holder->list, item);

	g_free(s);
}


