#include <string.h>
#include <glib.h>
#include "timer.h"

#include "utility.h"
#include "support.h"
#include "gconf.h"
#include "log.h"
#include "gst-recorder.h"

// Reference to global g_rec structure 
extern GstRec g_rec;

// Timer function call frequency (in seconds) 
#define TIMER_CALL_FREQ 4

void timer_func_stop();
gboolean timer_func_cb(gpointer userdata);
void timer_func_exit_cb(gpointer userdata);

void timer_func_check_values(gchar *timer_type, TimerRec *tr);

static gchar *hour_txt;
static gchar *min_txt;
static gchar *sec_txt;
static gchar *silence_txt;
static gchar *bytes_txt;

void timer_module_init() {
	// Translators: 
	// These strings h,m,s must start with a unique letter or sub-string. 
	// "h)our".
	hour_txt = _("h");  

	// "m)inute"
	min_txt =  _("m");

	// "s)econds"
	sec_txt =  _("s");

	// "silence" text
	silence_txt =  _("silence");

	// "bytes" text
	bytes_txt =  _("bytes");

	g_rec.timer_func_id = 0;
}

void timer_module_exit() {
	timer_func_stop();

    // Clean up/free static variables
    timer_func_cb((gpointer)0);
}

// ------------------------------------------------
// A simple lexer. Split a string to its tokens. 
// ------------------------------------------------
GSList *split_to_tokens(gchar *txt) { 
	// Split txt to tokens delimited by space.
	// Token can be a number, string. 
	// Return token GSlist 

	if (!txt) return NULL;

	#define MAX_LEN 64

	GSList *lst = NULL;

	// strdown
	gchar *s = g_utf8_strdown(txt, -1);
	glong len = g_utf8_strlen(s, MAX_LEN);

	gchar *end = g_utf8_offset_to_pointer(s, len - 1);
	gchar *p = s;
	
	while (p <= end)
	{
		gchar *tok = NULL;

		// gunichar unichar = g_utf8_get_char(p);

		if (*p == '\0') break;

		gchar *start = p;

		if (g_unichar_isspace(*p)) {
			p = g_utf8_next_char(p);
			continue;
		}

		// An integer or decimal number: 0-9, 0-9.0-9		
		while ((g_unichar_isdigit(*p) ||  g_utf8_strchr(p, 1, '.')) &&  p <= end) {
			// g_print("digit=%c\n", *p); 
			p = g_utf8_next_char(p);
		} 

		if (p > start) {
			// Copy the "numeric" token	
			tok = g_strndup(start, (gsize)(p - start));	

			// Put back the last char
			p = g_utf8_prev_char(p);
		}
		else { 
			// Must be a letter, character
			
			while (!g_unichar_isspace(*p) && !g_unichar_isdigit(*p) && p <= end) {
				p = g_utf8_next_char(p);
			} 

			if (p > start) {
				// Copy token
				tok = g_strndup(start, (gsize)(p - start));	
				// Put back the last char
				p = g_utf8_prev_char(p);
			}
		}

		// Got token?
		if (tok) {
			lst = g_slist_append(lst, tok);
		}

		// pos_ext = g_utf8_pointer_to_offset(path, pos);

		// Take next character
		p = g_utf8_next_char(p);
	}

	// Return GSlist
	return lst;
}

LR *lexer_init(gchar *txt) {
	// Init, create a new lexer
	LR *l = g_malloc0(sizeof(LR));
	l->txt = g_utf8_strdown(txt, -1);
	l->lst = split_to_tokens(l->txt);
	l->pos = 0;
	return l;	
}

void lexer_delete(LR **l) {
	// Delete lexer record
	if (!(l && *l)) return;
	LR *ll = *l;
	g_free(ll->txt);
	g_slist_foreach(ll->lst, (GFunc)g_free, NULL);
	g_slist_free(ll->lst);
	*l = NULL;
}

gchar *lexer_next_tok(LR *l) {
	// Get next token
	if (l->pos+1 > g_slist_length(l->lst)) return NULL;  

	gchar *txt = (gchar*)g_slist_nth_data(l->lst, l->pos); 
	l->pos += 1;		

	return txt;
}

void lexer_print(LR *l) {
	// Print lexer values
	if (!l) return;

	g_print("pos=%d txt=<%s>\n", l->pos, l->txt);

	GSList *item = l->lst;
	while (item) {
		gchar *txt = item->data;
		g_print("\t%s\n", txt);	
		item = g_slist_next(item);  
	}

} 

gboolean get_number(gchar *txt, gfloat *num) {
	if (!txt) return FALSE;

	// Return txt as number.
	// Test for validity.

	gchar *rest = NULL;
	*num  = (gfloat)g_strtod(txt, &rest);

	return !(strlen(txt) == strlen(rest));
}

gboolean starts_with(gchar *s, gchar *what) {
	// Test if string s starts with "what" (has that prefix) 
	
	if (!(s && what)) return FALSE;

	const gchar *p = g_strstr_len(s, -1, what);
	// The "s" starts with "what", at 1.st byte. 
	return (p == s);
} 

TimerRec *parse_timer_str(gchar *txt) {
	// Parse txt and return a pointer to TimerRec filled with values. 

    // TODO: Make this a UTF8 compliant.

    // Allocate new TimerRec structure.
    // hour=0.0, min=0.0, sec=0.0, other=0.0, label=NULL.
    TimerRec *tr = g_malloc0(sizeof(TimerRec)); 
    
	// Create a lexer
	LR *l = lexer_init(txt);

	gfloat last_val = 0;
	guint sep_count = 0; // count ":"s
	gfloat num = 0;

	while (1) {

		// Get token
		gchar *t = lexer_next_tok(l);

		// All tokens consumed?
		if (t == NULL) break;

		// Is numeric?
		gboolean is_num = get_number(t, &num);

        // Example strings:
		// 31 min			// "Stop after" 31 minutes.
		// 02:34:45			// "Stop at" or "Stop after" 02:34:45.
		// 2h 34 min 45 s   // Same.
		// silence 			// Stop at "silence", the dB level goes under ~10db.
		// silence 10 s		// Stop after 10 sec silence.
		// 3000000 bytes    // Stop when file size >= 3000000 bytes
		// 3000 KB			// Same.
		// 3.0 MB			// Sam.e
		// 3 GB				// Stop when file size >= 3 GB.

		if (is_num) {
			// Save it
			last_val = num;
				
			// Get next token
			t = lexer_next_tok(l);

			// Do not accept 2 numbers in a row
			gfloat tmp; 
			if (get_number(t, &tmp)) {
				t = NULL;
			}
		} 

		// The token t is?

		// "silence"
		if (starts_with(t, silence_txt) || starts_with(t, "silence")) {
            g_free(tr->label);
			tr->label = g_strdup(t);
		}
		// "bytes", "KB", "MB", "GB"
		else if (starts_with(t, "bytes") || starts_with(t, bytes_txt) || starts_with(t, "kb") || starts_with(t, "mb") || starts_with(t, "gb")) {
			tr->other = last_val;
            g_free(tr->label);
			tr->label = g_strdup(t);
		}
		// Clock time of format "01:20:30", "07 am" or "07 pm"
		else if (starts_with(t, ":") || 
                 t == NULL || 
                 !g_strcmp0(t, "pm") ||
                 !g_strcmp0(t, "am") ||
                 sep_count > 0/*last one*/) {

            // Make sure this is understood as clock time.
            g_free(tr->label);
			tr->label = g_strdup("clock time");

			if (sep_count == 0) {
				tr->hour = last_val;
			} else if (sep_count == 1) {
				tr->min = last_val;
			} else if (sep_count == 2) {
				tr->sec = last_val;
			}

            if (!g_strcmp0(t, "pm") && tr->hour <= 12) {
                // Convert to 24 hour clock.
                // Eg. "07:00 pm" becomes "19:00"
                tr->hour += 12.0;
                if (tr->hour > 24.0) tr->hour = 24.0;
            }

			sep_count += 1;
		}
        // Time period/duartion of form 1h 2m 3s (hours minutes seconds).

		// "1 h"  ("h", "hour" in both localized language and english)
		else if (starts_with(t, hour_txt) || starts_with(t, "h")) {
			tr->hour = last_val;
		}
		// "1 m"
		else if (starts_with(t, min_txt) || starts_with(t, "m")) {
			tr->min = last_val;
		}
		// "1 s"
		else if (starts_with(t, sec_txt) || starts_with(t, "s")) {
			tr->sec = last_val;
		}

		last_val = -1;
	}

	// Delete the lexer
	lexer_delete(&l);

	// Got we values at all?
	if (tr->hour == 0.0 && 
        tr->min == 0.0 && 
        tr->sec == 0.0 &&
        tr->other == 0.0 && 
        tr->label == NULL) {

        // Free tr 
        g_free(tr);
        return NULL;
    }

	// Normalize decimal values to real hours/minutes/seconds 
	// Eg. 2.5h means 2h 30minutes
	gdouble secs = tr->hour*3600.0 + tr->min*60.0 + tr->sec;

	// Convert back to hour/min/sec
	tr->hour = (guint)(secs / 3600);
	secs = secs - (tr->hour*3600);
	tr->min = (guint)(secs / 60);
	tr->sec = secs - (tr->min*60);

	return tr;
}

GList *split_to_string_list(gchar *str) {
    // Split the str on delimiters '|' and '!'.
    // Return a GList of strings.

    // TODO: Make this a UTF8 compliant.
    // See the example below.

    GList *list = NULL;
    if (!str) return list;

    // Split the str.
    // Note: This works on ASCII strings but probably not on UTF8 text.
    gchar **args = g_strsplit_set(str, "|!", -1);
    
    guint i =0;
    for (i=0; args[i]; i++) {
        list = g_list_append(list, g_strdup(args[i]));
    }

    // Free the (ASCII) string list
    g_strfreev(args);

    /*
    // TODO: Make this (split string) function UTF8 compatible.
    // An example:

    gchar *c = str;
    GString *s = NULL;

    guint count = 0;
    while (*c || count > 1024) {
        glong pos = g_utf8_pointer_to_offset(str, c);

        if (!s)    
            s = g_string_sized_new(64);

        if (*c == '|' || *c == '!')
		{
            g_string_append_c(s, '\0');
            list = g_list_append(list, g_string_free(s, FALSE));
            s = NULL;
		} else {
            g_string_append_unichar(s, *c);
        }

        c = g_utf8_find_next_char(c, NULL); 

        count += 1;
    }

    if (s)
        list = g_list_append(list, g_string_free(s, FALSE));
    */
    return list;
}

GList *parse_timer_str_many(gchar *str) {
    // Split str on '|' and '!', then parse/evaluate each sub string to TimerRec values.
    // Return a GList of TimerRec pointers.
    // 
    // Typical str is: "500 MB | silence | 1h 20min".

    // First, split str on '|' and '!'.
    GList *lst = split_to_string_list(str);
    GList *timer_val_lst = NULL;

    // Now parse each substring and create TimerRec structures.
    GList *item = g_list_first(lst);
    while (item) {
        // Get TimerRec structure
        TimerRec *tr = parse_timer_str((gchar*)item->data);
        if (tr) {
            // Add it to the list
            timer_val_lst = g_list_append(timer_val_lst, tr);
        }    
        item = g_list_next(item);
    }

    // Free the GList of string values
    g_list_foreach(lst, (GFunc)g_free, NULL);
    g_list_free(lst);

    // Return the TimerRec GList
    return timer_val_lst;
}

void free_timer_val_list(GList **lst) {
    // Free GList of TimerRec pointers/values
    if (!(lst && *lst)) return;

    GList *item = g_list_first(*lst);
    while (item) {
        TimerRec *tr = (TimerRec*)item->data;
        g_free(tr->label);
        g_free(tr);
        item->data = NULL;
        item = g_list_next(item);
    }

    g_list_free(*lst);
    *lst = NULL;
}

// --------------------------------------------------
// The actual timer function
// --------------------------------------------------

void timer_func_start() {
	// Start the timer function 
	g_rec.timer_func_id = g_timeout_add_seconds_full(G_PRIORITY_DEFAULT, TIMER_CALL_FREQ, (GSourceFunc)timer_func_cb, (gpointer)1/*!= 0*/, (GDestroyNotify)timer_func_exit_cb);
}

void timer_func_stop() {
	// Stop the timer funcion
	if (g_rec.timer_func_id > 0) {
		g_source_remove(g_rec.timer_func_id); 
		g_rec.timer_func_id = 0;
	}
}

void timer_func_exit_cb(gpointer userdata) {
	// Nothing to cleanup
	;
}

gboolean timer_func_cb(gpointer userdata) {
	// The actual timer function

    // "Stop at", "Start at"
	static gchar *timer_type = NULL;

    // Timer ON/OFF?
	static gboolean timer_active = FALSE;	

    // A GList of TimerRec pointers
    static GList *timer_values = NULL;

	static gint setting_counter = -1;

    // Clean up
    // If userdata == 0, clean up and exit.
    if (userdata == 0) {
        g_free(timer_type); timer_type = NULL;
        free_timer_val_list(&timer_values);
        timer_func_check_values(NULL, NULL);
        return FALSE;
    }

	// Are we recording?
	if (!gst_recorder_is_recording()) {
		goto LBL_1;
	}

    gint val = 0;
	// Timer settings changed?
	conf_get_int_value("timer_setting_counter", &val);

	if (val != setting_counter) {
		setting_counter = val;

        // Free old values
        g_free(timer_type);timer_type = NULL;
        free_timer_val_list(&timer_values);
    
        // Get new values
		conf_get_boolean_value("timer_active", &timer_active);
		conf_get_string_value("timer_type", &timer_type);

        gchar *value_str = NULL;
        conf_get_string_value("timer_value", &value_str);

		LOG_DEBUG("Timer settings changed:<%s>  <%s>  <%s>\n", (timer_active ? "timer ON" : "timer OFF"), timer_type, value_str);

        // Get GList of TimerRec values/pointers
        timer_values = parse_timer_str_many(value_str);
        g_free(value_str);

        if (g_list_length(timer_values) < 1)
            LOG_DEBUG("The timer has no valid values.\n");
        else
            LOG_DEBUG("The new timer values are:\n");
            
        GList *item = g_list_first(timer_values);
        while (item) {
            TimerRec *tr = (TimerRec*)item->data;
            if (tr) {} 
            LOG_DEBUG("\thour=%3.3f  min=%3.3f sec=%3.3f other=%3.3f  label=%s\n", tr->hour, tr->min, tr->sec, tr->other, tr->label);
            item = g_list_next(item);
        }
        
	}

	// Timer is ON/OFF?
	if (!timer_active) {
		goto LBL_1;
	}

    // For all TimerRec structures in the GList...
    GList *item = g_list_first(timer_values);
    while (item) {
        TimerRec *tr = (TimerRec*)item->data;

        // Check the timer values
        timer_func_check_values(timer_type, tr);

        // Next item
        item = g_list_next(item);
    }

  LBL_1:
	// Continue calling this function
	return TRUE;

}    

void timer_func_check_values(gchar *timer_type, TimerRec *tr) {

	#define DEF_SILENCE_DURATION 5 // in seconds 
	#define DEF_SILENCE_LEVEL_VALUE 0.16 // 0 - 1.0

    // Count seconds
	static gint silence_count = 0;

    // Called with NULL?
    if (!(timer_type && tr)) {
        // Reset static variables and return 
        silence_count = 0;
    }

	// Are we recording?
	if (!gst_recorder_is_recording()) {
		goto LBL_1;
	}

    guint i;
    gint64 stream_time = gst_get_stream_time(); // in seconds

    // "Start at" and "Stop at" have same meaning in the following tests. 
    // Actually we do not check these words at all.
    
    // Test for "silence"
    if ((starts_with(tr->label, silence_txt) || starts_with(tr->label, "silen"))) { // && stream_time > 9/*seconds*/ ) {
	    // Check for silence

        // Has user given a silence period?
        // We do not care about tr->hour here.
        gfloat seconds = tr->min*60.0 + tr->sec;
        if (seconds < 0.0)
            seconds = DEF_SILENCE_DURATION;

	    // Set data lock
	    rec_set_data_lock();

	    /// Get sum of level values
	    gdouble sum_level = 0.0;
	    for (i=0; i < MAX_LEVEL_VALUES; i++) {
		    sum_level += g_rec.level_buf[i].norm_peak;
	    }

	    // Free the lock
	    rec_free_data_lock();

	    // Average level value
	    gdouble avg = sum_level/(MAX_LEVEL_VALUES * 1.0);

        LOG_DEBUG("Timer testing for \"silence\": count in seconds=%ds,  stop after=%3.3fs silence,  avg. audio level=%3.3f\n", silence_count, seconds, avg);

	    if (avg < DEF_SILENCE_LEVEL_VALUE) {
		    // Count seconds
		    silence_count += TIMER_CALL_FREQ;

		    if (silence_count >= seconds) {
			    LOG_DEBUG("Timer testing for \"silence\": TRUE, STOPPING RECORDING.\n");

			    // Stop recording
			    gst_recorder_stop_and_reset();

			    // Reset silence counter
			    silence_count = 0;
		    }

	    } else {
			    // Reset silence counter
			    silence_count = 0;
	    }

	    goto LBL_1;
    }	


    // File size in "bytes", "KB", "MB", "GB"
    if (starts_with(tr->label, "bytes") || starts_with(tr->label, bytes_txt) || starts_with(tr->label, "kb") || 
	    starts_with(tr->label, "mb") || starts_with(tr->label, "gb")) {

	    // Get output filename
	    gchar *filename = gst_get_output_filename();

	    if (!filename) {
		    LOG_DEBUG("Timer testing filesize. Cannot get filename.\n");
		    goto LBL_1;
	    }

	    // Get file size
	    gdouble filesize = 1.0 * get_file_size(filename);
	    LOG_DEBUG("Timer testing for filesize: val=%3.3f unit=%s size filesize=%Ld bytes, filename=<%s>\n", tr->other, tr->label, (guint64)filesize, filename);

	    g_free(filename);

	    gboolean ret = FALSE;
	    if (starts_with(tr->label, "kb")) // x KB
		    ret = (filesize/1.0E3 >= tr->other); 
	    else if (starts_with(tr->label, "mb"))  // x MB
		    ret = (filesize/1.0E6 >= tr->other); 
	    else if (starts_with(tr->label, "gb")) // x GB
		    ret = (filesize/1.0E9 >= tr->other); 
	    else // "bytes"
		    ret = (filesize >= tr->other); 

	    if (ret) {
		    LOG_DEBUG("Timer testing for filesize: TRUE, STOPPING RECORDING.\n");

		    // Stop recording
		    gst_recorder_stop_and_reset();
	    }		
	    goto LBL_1;
    }

    // Else test time values like
    // 1 h 3 min 2s  (time duration)
    // 01:25:31  (exact clock time)

    // User has given time values?
    if (tr->hour <= 0.0 && tr->min <= 0.0 && tr->sec <= 0.0) {
	    goto LBL_1;		
    }

    // Get current time value	
    time_t t = time(NULL);
    struct tm *tmp;
    tmp = localtime(&t);

    // Is it a clock time of format "hh:mm:ss"?
    if (!g_strcmp0(tr->label, "clock time")) {
        // Test for exact clock time.

	    // Stop at: clock time
	    if ((tmp->tm_hour >= tr->hour) && 
	        (tmp->tm_min  >= tr->min) && 
	        (tmp->tm_sec  >= tr->sec)) {

		    LOG_DEBUG("Timer testing for clock time: TRUE, STOPPING RECORDING.\n");

		    // Stop recording
		    gst_recorder_stop_and_reset();
	    } 

    } else {
	    // "Stop after", stop after time duration.
	    // Test for duration.

	    if ((gdouble)stream_time >= tr->hour*3600.0 + tr->min*60.0 + tr->sec) {  

		    LOG_DEBUG("Timer testing for duration: TRUE, STOPPING RECORDING.\n");

		    // Stop recording
		    gst_recorder_stop_and_reset();
	    }
    }


  LBL_1:
   ;
}

