#include <glib.h>
#include <glib/gstdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#include "support.h" // _(x)
#include "utility.h"

gboolean is_file_writable(gchar *filename)
{
	GFileOutputStream *fstream;
	GError *error = NULL;
	GFile *file = g_file_new_for_path(filename);
	gboolean del = FALSE;

	if (g_file_test(filename, G_FILE_TEST_IS_REGULAR)) {
		fstream = g_file_append_to(file, G_FILE_CREATE_NONE, NULL, &error);
	} else {
		fstream = g_file_create(file, G_FILE_CREATE_NONE, NULL, &error);
		del = (!error);
	}

	gboolean ret = TRUE;

	if (error || !fstream) {
		ret = FALSE;
	} 

	if (error)
		g_error_free(error);
	g_object_unref(fstream);
	g_object_unref(file);

	if (del)
		g_remove(filename);

	return ret;
}

guint str_length(gchar *s, guint maxlen)
{
	if (!s) return 0;
	return g_utf8_strlen(s, maxlen);
}

guint str_length0(gchar *s)
{
	return str_length(s, MAX_PATH_LEN);
}

void split_filename2(gchar *path, gchar **filepath,  gchar **filebase)
{
	gchar *fileext;
	split_filename3(path, filepath,  filebase, &fileext);
	if (fileext)
	{
		/* TODO: Not very optimal code this... path is much longer than (filebase + '.' + fileext)
        */
		gchar *result = g_strdup(path);	
		g_sprintf(result, "%s.%s", *filebase, fileext);
		g_free(fileext);
		g_free(*filebase);
		*filebase = result;
	}
}

void split_filename3(gchar *path, gchar **filepath,  gchar **filebase, gchar **fileext)
{
	static gchar buf[MAX_PATH_LEN];

	*fileext = NULL;
	*filebase = NULL;
	*filepath = NULL;

	if (!path || str_length(path, MAX_PATH_LEN) < 1) return;

	gchar *fpath = NULL;
	gchar *fbase = NULL;
	gchar *fext  = NULL;

	glong pos_ext = -1;
	glong pos_base = -1;	

	glong path_len = str_length(path, MAX_PATH_LEN);

	gchar *pos = g_utf8_offset_to_pointer(path, path_len - 1);
	while (pos >= path)
	{
		gunichar unichar = g_utf8_get_char(pos);

		if (unichar == '.')
			pos_ext = g_utf8_pointer_to_offset(path, pos);
		else if (unichar == '/')
		{
			pos_base = g_utf8_pointer_to_offset(path, pos);
			break;
		}		
		pos = g_utf8_prev_char(pos);
	}

	glong len;

	/* Take the file name (basename) before file extension */
	if (pos_base > -1)
	{
		if (pos_ext > -1)
			len = pos_ext - pos_base -1;
		else
			len = path_len - pos_base -1;
	}
	else
	{
		if (pos_ext > -1)
			len = pos_ext;
		else
			len = path_len;
	}

	if (len > 0)
	{
		memset(buf, '\0', 1024);
		gchar *pos = g_utf8_offset_to_pointer(path, pos_base+1);
		g_utf8_strncpy(buf, pos, (gsize)len);
		fbase = g_strdup(buf);
	}

	/* Take the unused path before basename and file extension */
	if (pos_base > -1)
	{
		memset(buf, '\0', 1024);
		g_utf8_strncpy(buf, path, (gsize)pos_base+1);
		fpath = g_strdup(buf);
	}

	/* Take file extension without "." */
	if (pos_ext > -1)
	{
		len = path_len - pos_ext - 1;
		if (len > 0 )
		{
			memset(buf, '\0', 1024);
			gchar *pos = g_utf8_offset_to_pointer(path, pos_ext+1);
			g_utf8_strncpy(buf, pos, (gsize)len);
			fext = g_strdup(buf);
		}
	}

	*filepath = fpath;
	*filebase = fbase;
	*fileext = fext;
}

gchar *replace_str(gchar *what, gchar *width) {
	// Replace space and tab with '_'
	// g_strdelimit(pbasename, " \t", '_');

	return NULL;
} 

gboolean paths_are_equal(gchar *path1, gchar *path2) {
	// Add trailing /
	gchar *p1 = g_build_filename(path1, "/", NULL);
	gchar *p2 = g_build_filename(path2, "/", NULL);

	gboolean ret = !g_strcmp0(p1, p2);
	
	g_free(p1);
	g_free(p2);

	return ret;
}

gchar *format_file_size(guint64 fsize) {
	gint div = 1;
	gchar *label;

	if (fsize > 1E9) {
		div = 1E9;
		label = "GB";	
	} else if (fsize > 1E6) {
		div = 1E6;
		label = "MB";	
	} else if (fsize > 1E3) {
		div = 1E3;
		label = "KB";	
	} else {
		div = 1;
		label = "bytes";
	}

	gchar *txt = g_strdup_printf("%02.1F %s", ((gdouble)fsize / (gdouble)div), label);  

	// The caller should g_free() this value
	return txt;
} 

gchar *substitute_time_and_date_pattern(gchar *pattern) {
	// Typical pattern is: "%C%y-%m-%d-%R:%S"
	// See: http://linux.die.net/man/3/strftime
	time_t t;
	struct tm *tmp;
	t = time(NULL);
	tmp = localtime(&t);
	if (tmp == NULL) {
		return NULL;
	}

	static gchar buf[128];

	strftime(buf, 128, pattern, tmp);

	// Caller should g_free() this value
	return g_strdup(buf);
}

guint64 get_file_size(gchar *filename) {
	// Get the file size
	struct stat fstat;
	if (!stat(filename, &fstat)) {
		return (guint64)fstat.st_size;
	}
	return 0L;
}

gint messagebox_error(const gchar *message, GtkWidget *window)
{
	/* Show dialog box with error message and wait for user's response. */
	GtkWidget *dialog;
	GtkResponseType answer;
	
	dialog = gtk_message_dialog_new((window ? GTK_WINDOW(window) : NULL), 
		                             GTK_DIALOG_DESTROY_WITH_PARENT, 
		                             GTK_MESSAGE_ERROR,
		                             GTK_BUTTONS_CLOSE,
		                             message, NULL);

	gtk_window_set_title(GTK_WINDOW(dialog), _("Audio Recorder Application"));

	answer = gtk_dialog_run(GTK_DIALOG(dialog));
	/* answer not used */

	gtk_widget_destroy(dialog);

	return TRUE;
}

GPid exec_command_async(gchar **argv, GError **error)
{	
	/* Start, spawn a child process and run the command given by argv list.
       The process is asynchronous and returns immediately back to the caller.
	   This function returns PID of the child process or -1 if errors.
	*/

	GPid child_pid;

	if (error) *error = NULL;

	if (!g_spawn_async(NULL/*inherit parent*/, argv, NULL/*inherit parent*/, G_SPAWN_LEAVE_DESCRIPTORS_OPEN | G_SPAWN_SEARCH_PATH, 
                       NULL/*no setup function*/, NULL/*no user data*/, &child_pid, error))
	{
        g_printerr(_("Error. Cannot start process %s.\n%s.\n"), argv[0], (*error)->message);
		return -1;
	}

	/* The child process is running. Return its pid */
	return child_pid;
}

#if 0 
void run_program_for_file(void *user_data)
{
	if (!user_data) return (void*)0;

	gchar *filename = (gchar*)user_data;

	/* Init Gnome's VFS.
	   EDIT: I moved this to main.c.
	   gnome_vfs_init(); 
	*/

	gchar *mime_type = gnome_vfs_get_mime_type(filename);
	if (!mime_type) goto l_end;

	GnomeVFSMimeApplication *app = gnome_vfs_mime_get_default_application(mime_type);
	if (!app) goto l_end;

	GList *uris = g_list_prepend(NULL, filename);
	gnome_vfs_mime_application_launch(app, uris);
	g_list_free(uris);/* We will free filename later */

	gnome_vfs_mime_application_free(app);

 l_end:

	/* TODO: 
	   Check if we can shut this down.
	   gnome_vfs_shutdown();
	*/

	/* Free the mime-type and filename */
	g_free(mime_type);
	g_free(filename);

	/* Exit ok */
	g_thread_exit(0);

	return (void*)0;
}
#endif

